from collections.abc import Callable, Mapping
from typing import Any, Literal, overload

import array_api_extra as xpx
import numpy as np
from array_api._2024_12 import Array, ArrayNamespaceFull
from scipy.special import roots_jacobi

from ._coordinates import (
    BranchingType,
    SphericalCoordinates,
    TCartesian,
    TSpherical,
    get_child,
)


def roots(
    c: SphericalCoordinates[TSpherical, TCartesian],
    n: int,
    *,
    expand_dims_x: bool,
    expand_dims_w: bool = False,
    device: Any | None = None,
    dtype: Any | None = None,
    xp: ArrayNamespaceFull,
) -> tuple[Mapping[TSpherical, Array], Mapping[TSpherical, Array]]:
    r"""
    Gauss-Jacobi quadrature roots and weights.

    .. math::
        \int_\mathbb{S}^{d-1} f d\omega^{d-1} \approx
        \sum_{(\theta_1, w_1)} w_1 \cdots \sum_{(\theta_{d-1}, w_{d-1})} w_{d-1}
        f(\theta_1, \ldots, \theta_{d-1})

    Parameters
    ----------
    n : int
        The number of roots.
    expand_dims_x : bool
        Whether to expand dimensions of the roots, by default False
    expand_dims_w : bool, optional
        Whether to expand dimensions of the weights, by default False
    device : Any, optional
        The device, by default None
    dtype : Any, optional
        The data type, by default None

    Returns
    -------
    tuple[Mapping[TSpherical, Array], Mapping[TSpherical, Array]]
        roots and weights

    Raises
    ------
    ValueError
        If the branching type is invalid.

    Example
    -------
    >>> from array_api_compat import numpy as np
    >>> import ultrasphere as us
    >>> c = us.create_spherical()
    >>> xs, ws = us.roots(c, 5, expand_dims_x=True, expand_dims_w=True, xp=np)
    >>> xs
    {'theta': array([[2.7049577 ],
           [2.13941585],
           [1.57079633],
           [1.0021768 ],
           [0.43663495]]), 'phi': array([[0.        , 0.62831853, 1.25663706, 1.88495559, 2.51327412,
            3.14159265, 3.76991118, 4.39822972, 5.02654825, 5.65486678]])}
    >>> ws
    {'theta': array([[0.23692689],
           [0.47862867],
           [0.56888889],
           [0.47862867],
           [0.23692689]]), 'phi': array([[0.62831853, 0.62831853, 0.62831853, 0.62831853, 0.62831853,
            0.62831853, 0.62831853, 0.62831853, 0.62831853, 0.62831853]])}

    """
    xs = {}
    ws = {}
    for i, node in enumerate(c.s_nodes):
        branching_type = c.branching_types[node]
        if branching_type == BranchingType.A:
            x = xp.arange(2 * n, device=device, dtype=dtype) * xp.pi / n
            w = xp.ones(2 * n, device=device, dtype=dtype) * xp.pi / n
        elif branching_type == BranchingType.B:
            s_beta = c.S[get_child(c.G, node, "sin")]
            beta = s_beta / 2
            x, w = roots_jacobi(n, beta, beta)
            x = np.acos(x)
        elif branching_type == BranchingType.BP:
            s_alpha = c.S[get_child(c.G, node, "cos")]
            alpha = s_alpha / 2
            x, w = roots_jacobi(n, alpha, alpha)
            x = np.asin(x)
        elif branching_type == BranchingType.C:
            s_alpha = c.S[get_child(c.G, node, "cos")]
            s_beta = c.S[get_child(c.G, node, "sin")]
            alpha = s_alpha / 2
            beta = s_beta / 2
            x, w = roots_jacobi(n, alpha, beta)
            w /= 2 ** (alpha + beta + 2)
            x = np.acos(x) / 2
        else:
            raise ValueError(f"Invalid branching type {branching_type}.")
        x = xp.asarray(x, device=device, dtype=dtype)
        w = xp.asarray(w, device=device, dtype=dtype)
        if expand_dims_x:
            x = x[(None,) * i + (slice(None),) + (None,) * (c.s_ndim - i - 1)]
        if expand_dims_w:
            w = w[(None,) * i + (slice(None),) + (None,) * (c.s_ndim - i - 1)]
        xs[node] = x
        ws[node] = w
    return xs, ws


@overload
def integrate(
    c: SphericalCoordinates[TSpherical, TCartesian],
    f: (
        Callable[
            [Mapping[TSpherical, Array]],
            Mapping[TSpherical, Array],
        ]
        | Mapping[TSpherical, Array]
    ),
    does_f_support_separation_of_variables: Literal[True],
    n: int,
    *,
    xp: ArrayNamespaceFull,
    device: Any | None = None,
    dtype: Any | None = None,
) -> Mapping[TSpherical, Array]: ...


@overload
def integrate(
    c: SphericalCoordinates[TSpherical, TCartesian],
    f: (
        Callable[
            [Mapping[TSpherical, Array]],
            Array,
        ]
        | Array
    ),
    does_f_support_separation_of_variables: Literal[False],
    n: int,
    *,
    xp: ArrayNamespaceFull,
    device: Any | None = None,
    dtype: Any | None = None,
) -> Array: ...


def integrate(
    c: SphericalCoordinates[TSpherical, TCartesian],
    f: (
        Callable[
            [Mapping[TSpherical, Array]],
            Mapping[TSpherical, Array] | Array,
        ]
        | Mapping[TSpherical, Array]
        | Array
    ),
    does_f_support_separation_of_variables: bool,
    n: int,
    *,
    xp: ArrayNamespaceFull,
    device: Any | None = None,
    dtype: Any | None = None,
) -> Array | Mapping[TSpherical, Array]:
    r"""
    Integrate the function over the hypersphere.

    .. math::
        \int_{\mathbb{S}^{d-1}} f d\omega^{d-1}

    Parameters
    ----------
    f : Callable[ [Mapping[TSpherical, Array]], Mapping[TSpherical, Array] | Array, ] | Mapping[TSpherical, Array] | Array # noqa: E501
        The function to integrate or the values of the function.

        If mapping, the separated parts of the function for each spherical coordinate.

        If mapping, the shapes do not need to be broadcastable.

        If function, if does_f_support_separation_of_variables is True,
        1D array of integration points are passed,
        and extra axis should be added to the last dimension.

        If function, if does_f_support_separation_of_variables is False,
        ``c.s_ndim``-D array of integration points are passed,
        and extra axis should be added to the last dimension.
    does_f_support_separation_of_variables : bool
        Whether the function supports separation of variables.
        This could significantly reduce the computational cost.
    n : int
        The number of roots.
    device : Any, optional
        The device, by default None
    dtype : Any, optional
        The data type, by default None

    Returns
    -------
    Array | Mapping[TSpherical, Array]
        The integrated value.
        Has the same shape as the return values of f or the values of f.

    Example
    -------
    >>> from array_api_compat import numpy as np
    >>> import ultrasphere as us
    >>> c = us.create_spherical()
    >>> f = lambda spherical: spherical["theta"] ** 2 * spherical["phi"]
    >>> np.round(us.integrate(
    ...     c,
    ...     f,
    ...     False, # does not support separation of variables
    ...     10, # number of quadrature points
    ...     xp=np # the array namespace
    ... ), 5)
    np.float64(110.02621)

    """
    xs, ws = roots(
        c,
        n,
        device=device,
        dtype=dtype,
        expand_dims_x=not does_f_support_separation_of_variables,
        xp=xp,
    )
    if isinstance(f, Callable):  # type: ignore
        try:
            val = f(xs)  # type: ignore
        except Exception as e:
            raise RuntimeError(f"Error occurred while evaluating {f=}") from e
    else:
        val = f

    # in case f(theta1, ...) = f_1(theta1) * f_2(theta2) * ...
    if isinstance(val, Mapping):
        result = {}
        for node in c.s_nodes:
            value = val[node]
            # supports vectorized function
            # axis=0 because in sph_harm
            # we add axis to the last dimension
            # theta(node),u1,...,uM
            xpx.broadcast_shapes(value.shape[:1], ws[node].shape)
            w = xp.reshape(ws[node], (-1,) + (1,) * (value.ndim - 1))
            if value.shape[0] == 1:
                result[node] = value[0, ...] * xp.sum(w)
            else:
                result[node] = xp.vecdot(w, value, axis=0)
        # we don't know how to einsum the result
        return result
    if val.ndim < c.s_ndim:
        raise ValueError(
            f"The dimension of the return value of f should be at least {c.s_ndim}, got {val.ndim}."
        )
    xpx.broadcast_shapes(
        val.shape[: c.s_ndim],
        xpx.broadcast_shapes(*(xs[node].shape for node in c.s_nodes)),
    )
    # theta1,...,thetaN,u1,...,uM\
    for node in c.s_nodes:
        w = ws[node]
        if val.shape[0] == 1:
            val = val[0, ...] * xp.sum(w)
        else:
            val = xp.vecdot(w[(slice(None),) + (None,) * (val.ndim - 1)], val, axis=0)
    return val



from collections.abc import Mapping, Sequence
from typing import Any, Literal, overload

import numpy as np
from array_api._2024_12 import Array, ArrayNamespaceFull
from numpy._typing import NDArray

from ultrasphere._coordinates import (
    BranchingType,
    SphericalCoordinates,
    TCartesian,
    TSpherical,
)


def _random_sphere(
    shape: Sequence[int],
    dim: int,
    *,
    surface: bool = False,
    rng: np.random.Generator | None = None,
) -> NDArray[np.float64]:
    r"""
    Generate random points in a unit ball / sphere.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the output array.
    dim : int
        The dimension of the hypersphere.
    surface : bool, optional
        Whether to generate points on the surface of the hypersphere,
        by default False.
    rng : np.random.Generator | None, optional
        The random number generator, by default None.

    Returns
    -------
    NDArray[np.float64]
        The generated points.

    References
    ----------
    Barthe, F., Guédon, O., Mendelson, S., & Naor, A. (2005).
    A probabilistic approach to the geometry of the ? P n -ball.
    The Annals of Probability, 33. https://doi.org/10.1214/009117904000000874

    """
    rng = np.random.default_rng() if rng is None else rng
    g = rng.normal(loc=0, scale=np.sqrt(1 / 2), size=(dim, *shape))
    if surface:
        result = g / np.linalg.vector_norm(g, axis=0)
    else:
        z = rng.exponential(scale=1, size=shape)[None, ...]
        result = g / np.sqrt(np.sum(g**2, axis=0, keepdims=True) + z)
    return np.nan_to_num(result, nan=0.0)  # not sure if nan_to_num is necessary


@overload
def random_ball(
    c: SphericalCoordinates[TSpherical, TCartesian],
    *,
    shape: Sequence[int],
    xp: ArrayNamespaceFull,
    device: Any | None = ...,
    dtype: Any | None = ...,
    type: Literal["uniform"] = ...,
    rng: np.random.Generator | None = ...,
    surface: bool = ...,
) -> Array: ...
@overload
def random_ball(
    c: SphericalCoordinates[TSpherical, TCartesian],
    *,
    shape: Sequence[int],
    xp: ArrayNamespaceFull,
    device: Any | None = ...,
    dtype: Any | None = ...,
    type: Literal["spherical"] = ...,
    rng: np.random.Generator | None = ...,
    surface: Literal[False] = ...,
) -> Mapping[TSpherical | Literal["r"], Array]: ...
@overload
def random_ball(
    c: SphericalCoordinates[TSpherical, TCartesian],
    *,
    shape: Sequence[int],
    xp: ArrayNamespaceFull,
    device: Any | None = ...,
    dtype: Any | None = ...,
    type: Literal["spherical"] = ...,
    rng: np.random.Generator | None = ...,
    surface: Literal[True] = ...,
) -> Mapping[TSpherical, Array]: ...
def random_ball(
    c: SphericalCoordinates[TSpherical, TCartesian],
    *,
    shape: Sequence[int],
    xp: ArrayNamespaceFull,
    device: Any | None = None,
    dtype: Any | None = None,
    type: Literal["uniform", "spherical"] = "uniform",
    rng: np.random.Generator | None = None,
    surface: bool = False,
) -> Array | Mapping[TSpherical | Literal["r"], Array] | Mapping[TSpherical, Array]:
    r"""
    Generate random points in the unit ball / sphere.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the random points.
    device : Any | None, optional
        The device, by default None
    dtype : Any | None, optional
        The dtype, by default None
    type : Literal["uniform", "spherical"], optional
        The type of the random points, by default "uniform"
        If "uniform", the random points are uniformly distributed on the sphere.
        If "spherical", each spherical coordinate (and radius
        if surface=False) is uniformly distributed.
    rng : np.random.Generator | None, optional
        The random number generator, by default None
    surface : bool, optional
        Whether to generate random points on the surface of the sphere
        or inside the sphere, by default False

    Returns
    -------
    Mapping[TSpherical, Array] | Array | Mapping[TSpherical | Literal["r"], Array]
        The random points.

    References
    ----------
    Barthe, F., Guédon, O., Mendelson, S., & Naor, A. (2005).
    A probabilistic approach to the geometry of the ? P n -ball.
    The Annals of Probability, 33. https://doi.org/10.1214/009117904000000874

    Examples
    --------
    >>> from array_api_compat import numpy as xp
    >>> import ultrasphere as us
    >>> c = us.create_standard(4)
    >>> rng = np.random.default_rng(0)
    >>> points_ball = random_ball(c, shape=(2,), xp=xp, rng=rng)
    >>> points_ball
    array([[ 0.04083225, -0.08099814],
           [ 0.20798418,  0.06431795],
           [-0.17396442,  0.22170665],
           [ 0.4234881 ,  0.58068866],
           [-0.22854562, -0.77587443]])
    >>> xp.linalg.vector_norm(points_ball, axis=0)
    array([0.55386242, 0.99951577])
    >>> points_sphere = random_ball(c, shape=(2,), xp=xp, rng=rng, surface=True)
    >>> points_sphere
    array([[-0.85238384, -0.11470655],
           [-0.45676572, -0.38390797],
           [-0.19953179, -0.16582762],
           [ 0.15090863,  0.54656157],
           [-0.04712233,  0.71639984]])
    >>> xp.linalg.vector_norm(points_sphere, axis=0)
    array([1., 1.])

    """
    rng = np.random.default_rng() if rng is None else rng
    if type == "uniform":
        return xp.asarray(
            _random_sphere(shape, dim=c.c_ndim, surface=surface, rng=rng),
            device=device,
            dtype=dtype,
        )
    elif type == "spherical":
        d = {
            BranchingType.A: (0, 2 * xp.pi),
            BranchingType.B: (0, xp.pi),
            BranchingType.BP: (-xp.pi / 2, xp.pi / 2),
            BranchingType.C: (0, xp.pi / 2),
        }
        result: dict[TSpherical | Literal["r"], Array] = {}
        for node in c.s_nodes:
            low, high = d[c.branching_types[node]]
            result[node] = xp.asarray(
                rng.uniform(low=low, high=high, size=shape), device=device, dtype=dtype
            )
        if not surface:
            result["r"] = xp.asarray(
                rng.uniform(low=0, high=1, size=shape), device=device, dtype=dtype
            )
        return result
    else:
        raise ValueError(f"Invalid type {type}.")



import warnings
from typing import Literal, TypeVar

import array_api_extra as xpx
from array_api._2024_12 import Array
from array_api_compat import array_namespace

from ultrasphere.special._bessel import shn1, sjv

TArray = TypeVar("TArray", bound=Array)


def potential_coef[TArray: Array](
    n: TArray,
    d: TArray,
    k: TArray,
    /,
    *,
    x_abs: TArray,
    y_abs: TArray,
    x_abs_derivative: bool | None = None,
    y_abs_derivative: bool | None = None,
    derivative: Literal["S", "D", "D*", "N"] | None = None,
    limit: Literal[False, "x_larger", "y_larger", "warn"] = "warn",
    for_func: Literal["harmonics", "solution"] = "harmonics",
) -> Array:
    r"""
    The coefficients for layer potentials.

    The coefficients for single-layer or double-layer potential
    for hyperspherical harmonics with homogeneous degree
    (maximum quantum number) n.

    y is the integral variable.

    .. math::
        r \mathbb{S}^{d-1} :=
        \{ x \in \mathbb{R}^d : \left|x\right| = r \}

    .. math::
        \forall d \in \mathbb{N} \setminus \{1\}.
        \forall x_a, y_a \in (0, \infty).
        \forall T \in \{S_{y_a \mathbb{S}^{d-1}}, D_{y_a \mathbb{S}^{d-1}}, D^*_{y_a \mathbb{S}^{d-1}}, N_{y_a \mathbb{S}^{d-1}}\}. \\
        \forall x \in x_a \mathbb{S}^{d-1}.
        \forall n \in \mathbb{N}_0.
        \forall Y_n \in \mathcal{H}(\mathbb{S}^{d-1}). \\
        T Y_n \left(\frac{x}{x_a}\right) = \text{potential_coef}(T) Y_n \left(\frac{x}{x_a}\right)


    Parameters
    ----------
    n : TArray
        The homogeneous degree of the hyperspherical harmonics.
        (maximum quantum number)
    d : TArray
        The dimension of the hypersphere.
    k : TArray
        The wavenumber.
    x_abs : TArray
        The distance from the origin O to the point x.
    y_abs : TArray
        The radius of the hypersphere.
    x_abs_derivative : bool | None, optional
        Whether the derivative is taken with respect to x_abs,
        by default False.
    y_abs_derivative : bool | None, optional
        Whether the derivative is taken with respect to y_abs,
        by default False.
    derivative : Literal["S", "D", "D*", "N"] | None, optional
        The shorthand for the derivative.
        Note that the integral variable is y.

        - "S" <=> x_abs_derivative = False, y_abs_derivative = False
        - "D" <=> x_abs_derivative = False, y_abs_derivative = True
        - "D*" <=> x_abs_derivative = True, y_abs_derivative = False
        - "N" <=> x_abs_derivative = True, y_abs_derivative = True
    limit : Literal[False, "x_larger", "y_larger"], optional
        Whether to return the directional derivative
        of the potential with respect to x
        to the x/|x| direction, by default False.
    for_func : Literal["harmonics", "solution"], optional
        Whether the coefficient is for the harmonics
        or the suitable (singular if outer, regular if inner)
        elementary solution, by default "harmonics".

    Returns
    -------
    Array
        The coefficient for the potential
        integrated by y.

    References
    ----------
    McLean, W. (2000). Strongly Elliptic Systems and
    Boundary Integral Equations. p.285

    """
    xp = array_namespace(n, d, k, x_abs, y_abs)
    if x_abs_derivative is None and y_abs_derivative is None:
        x_abs_derivative = derivative in ["D*", "N"]
        y_abs_derivative = derivative in ["D", "N"]
    elif x_abs_derivative is not None and y_abs_derivative is not None:
        pass
    else:
        raise ValueError(
            f"Both {x_abs_derivative=} and {y_abs_derivative=} must be None or not None."
        )
    # p.285
    inner = xp.where(
        x_abs < y_abs,
        shn1(n, d, k * y_abs, derivative=y_abs_derivative)
        * (
            sjv(n, d, k * x_abs, derivative=x_abs_derivative)
            if for_func == "harmonics"
            else 1
        ),
        sjv(n, d, k * y_abs, derivative=y_abs_derivative)
        * (
            shn1(n, d, k * x_abs, derivative=x_abs_derivative)
            if for_func == "harmonics"
            else 1
        ),
    )
    derivative_count = int(x_abs_derivative) + int(y_abs_derivative)
    result = (k) ** derivative_count * 1j * (k ** (d - 2)) * (y_abs ** (d - 1)) * inner
    if derivative_count == 1:
        if for_func == "solution" and limit in ["x_larger", "y_larger"]:
            raise NotImplementedError()
        addition = None
        if limit == "x_larger":
            if x_abs_derivative:
                # outer D^*
                addition = 1 / 2
            if y_abs_derivative:
                # outer D
                addition = -1 / 2
        elif limit == "y_larger":
            if x_abs_derivative:
                # inner D^*
                addition = -1 / 2
            if y_abs_derivative:
                # inner D
                addition = 1 / 2
        elif limit == "warn":
            if xp.any(xpx.isclose(x_abs, y_abs, rtol=1e-4, atol=0)):
                warnings.warn(
                    "As x_abs is close to y_abs "
                    "and requested to calculate D or D*, "
                    "which is not continuous at x_abs=y_abs, "
                    "it might be possible that what you really want "
                    "are the limit values."
                    "To calculate the limit values, set limit="
                    "'x_larger' or 'y_larger', depending on the "
                    "limit direction you want to take."
                    "To suppress this warning, set limit=False.",
                    RuntimeWarning,
                    stacklevel=2,
                )
        elif limit is False:
            pass
        else:
            raise ValueError(f"Invalid limit: {limit}")
        if addition is not None:
            result += addition
    return result


def fundamental_solution(
    d: TArray,
    z: TArray,
    k: TArray,
    derivative: bool = False,
) -> TArray:
    """
    Fundamental solution of the Helmholtz equation in d dimensions.

    Parameters
    ----------
    d : TArray
        The dimension of the space.
    z : TArray
        The argument of the fundamental solution of shape (..., d (coordinates)).
    k : TArray
        The wave number.
    derivative : bool, optional
        Whether to compute the derivative of the fundamental solution, by default False

    Returns
    -------
    TArray
        The fundamental solution of the Helmholtz equation of shape (...,).

    """
    xp = array_namespace(d, z)
    coef = k ** (d - 2) * 1j / (2 * (2 * xp.pi) ** ((d - 1) / 2))
    return coef * shn1(
        xp.asarray(0, device=z.device, dtype=xp.int32),
        d,
        k * xp.linalg.vector_norm(z, axis=-1),
        derivative,
    )



import math
from collections.abc import Callable, Mapping
from typing import Any, Literal

import array_api_extra as xpx
import pytest
from array_api._2024_12 import Array, ArrayNamespaceFull

from ultrasphere._coordinates import SphericalCoordinates, TCartesian, TSpherical
from ultrasphere._creation import (
    create_hopf,
    create_random,
    create_spherical,
    create_standard,
)
from ultrasphere._integral import integrate


@pytest.mark.parametrize("n", [4, 8, 16])
@pytest.mark.parametrize(
    "f, expected", [(lambda s: 1, 4 * math.pi), (lambda s: 1j, 4j * math.pi)]
)
def test_sphere_surface_integrate(
    f: Callable[[Mapping[Literal["theta", "phi"], Array]], Array],
    n: int,
    expected: float,
    xp: ArrayNamespaceFull,
) -> None:
    def f2(s):
        return xp.asarray(f(s)) * xp.ones_like(s["theta"])

    c = create_spherical()
    assert integrate(
        c, f2, does_f_support_separation_of_variables=False, n=n, xp=xp
    ).item() == pytest.approx(expected, rel=1e-2)


@pytest.mark.parametrize("r", [1, 3])
@pytest.mark.parametrize("n", [4, 8, 16])
@pytest.mark.parametrize(
    "c",
    [
        (create_spherical()),
        (create_standard(3)),
        (create_hopf(2)),
    ],
)
@pytest.mark.parametrize("concat", [True, False])
def test_integrate(
    c: SphericalCoordinates[TSpherical, TCartesian],
    n: int,
    concat: bool,
    r: float,
    xp: ArrayNamespaceFull,
) -> None:
    # surface integral (area) of the sphere
    def f(s: Mapping[TSpherical, Array]) -> Array:
        if concat:
            return xp.asarray(r**c.s_ndim) * xp.ones_like(next(iter(s.values())))
        else:
            return {k: xp.asarray(r) * xp.ones_like(s[k]) for k in c.s_nodes}

    actual = integrate(
        c,
        f,
        does_f_support_separation_of_variables=not concat,  # type: ignore
        n=n,
        xp=xp,
    )
    if not concat:
        actual = xp.prod(xp.stack(list(actual.values())))
    expected = c.surface_area(r)
    assert actual.item() == pytest.approx(expected, rel=1e-2)


@pytest.mark.parametrize("n", [3, 4, 5])
def test_integrate_match(n: int, xp: ArrayNamespaceFull) -> None:
    cs: list[SphericalCoordinates[Any, Any]] = [create_random(n - 1) for _ in range(4)]
    k = xp.random.random_uniform(low=0, high=1, shape=(n,))

    def create_f(
        c: SphericalCoordinates[TSpherical, TCartesian],
    ) -> Callable[[Mapping[TSpherical, Array]], Array]:
        def f(s: Mapping[TSpherical, Array]) -> Array:
            x = c.to_cartesian(s, as_array=True)
            return xp.einsum("v,v...->...", xp.astype(k, x.dtype), x)

        return f

    actual = [
        integrate(
            c, create_f(c), does_f_support_separation_of_variables=False, n=8, xp=xp
        )
        for c in cs
    ]
    assert xp.all(xpx.isclose(xp.asarray(actual), actual[0], rtol=1e-3, atol=1e-3))



__version__ = "2.0.3"
from ._coordinates import BranchingType, SphericalCoordinates, get_child, get_parent
from ._creation import (
    create_from_branching_types,
    create_hopf,
    create_polar,
    create_random,
    create_spherical,
    create_standard,
    create_standard_prime,
)
from ._draw import draw
from ._integral import integrate, roots
from ._random import random_ball
from .special import (
    fundamental_solution,
    potential_coef,
    shn1,
    shn2,
    sjv,
    syv,
    szv,
)

__all__ = [
    "BranchingType",
    "SphericalCoordinates",
    "create_from_branching_types",
    "create_hopf",
    "create_polar",
    "create_random",
    "create_random",
    "create_spherical",
    "create_standard",
    "create_standard_prime",
    "draw",
    "fundamental_solution",
    "get_child",
    "get_parent",
    "integrate",
    "potential_coef",
    "random_ball",
    "roots",
    "shn1",
    "shn2",
    "sjv",
    "syv",
    "szv",
]



from ._bessel import shn1, shn2, sjv, syv, szv
from ._helmholtz import fundamental_solution, potential_coef

__all__ = [
    "fundamental_solution",
    "potential_coef",
    "shn1",
    "shn2",
    "sjv",
    "syv",
    "szv",
]



from typing import Literal

import array_api_extra as xpx
import numpy as np
import pytest
from array_api._2024_12 import ArrayNamespaceFull
from scipy.special import hankel1, hankel2, jv, jvp, spherical_jn, spherical_yn, yv, yvp

from ultrasphere.special._bessel import shn1, shn2, sjv, syv


@pytest.mark.parametrize("derivative", [True, False])
@pytest.mark.parametrize("d", [2, 3])
@pytest.mark.parametrize("type", ["j", "y", "h1", "h2"])
def test_sjyn(
    d: Literal[2, 3],
    type: Literal["j", "y", "h1", "h2"],
    derivative: bool,
    xp: ArrayNamespaceFull,
) -> None:
    n = np.random.randint(5, size=(10, 10))
    z = np.random.random((10, 10))
    if type == "j":
        if d == 2:
            expected = np.sqrt(np.pi / 2) * (jvp(n, z) if derivative else jv(n, z))
        elif d == 3:
            expected = spherical_jn(n, z, derivative=derivative)
        else:
            raise ValueError("d must be 2 or 3")
        expected = xp.asarray(expected)
        actual = sjv(xp.asarray(n), xp.asarray(d), xp.asarray(z), derivative=derivative)
    elif type == "y":
        if d == 2:
            expected = np.sqrt(np.pi / 2) * (yvp(n, z) if derivative else yv(n, z))
        elif d == 3:
            expected = spherical_yn(n, z, derivative=derivative)
        else:
            raise ValueError("d must be 2 or 3")
        expected = xp.asarray(expected)
        actual = syv(xp.asarray(n), xp.asarray(d), xp.asarray(z), derivative=derivative)
    elif type == "h1":
        if derivative:
            pytest.skip("derivative of hankel1 not implemented")
        if d == 2:
            expected = np.sqrt(np.pi / 2) * hankel1(n, z)
        elif d == 3:
            expected = spherical_jn(n, z) + 1j * spherical_yn(n, z)
        expected = xp.asarray(expected)
        actual = shn1(xp.asarray(n), xp.asarray(d), xp.asarray(z))
    elif type == "h2":
        if derivative:
            pytest.skip("derivative of hankel2 not implemented")
        if d == 2:
            expected = np.sqrt(np.pi / 2) * hankel2(n, z)
        elif d == 3:
            expected = spherical_jn(n, z) - 1j * spherical_yn(n, z)
        expected = xp.asarray(expected)
        actual = shn2(xp.asarray(n), xp.asarray(d), xp.asarray(z))
    else:
        raise ValueError("type must be 'j' or 'y'")
    assert xp.all(xpx.isclose(actual, expected, rtol=1e-6, atol=1e-6))



from typing import Literal

import array_api_extra as xpx
import pytest
from array_api._2024_12 import Array, ArrayNamespaceFull
from array_api_compat import array_namespace

from ultrasphere._coordinates import SphericalCoordinates, TCartesian, TSpherical
from ultrasphere._creation import create_hopf, create_random, create_spherical


def to_cartesian(*, r: Array, theta: Array, phi: Array) -> tuple[Array, Array, Array]:
    xp = array_namespace(r, theta, phi)
    rsin = r * xp.sin(theta)
    x = rsin * xp.cos(phi)
    y = rsin * xp.sin(phi)
    z = r * xp.cos(theta)
    return x, y, z


@pytest.mark.parametrize(
    "shape",
    [
        (1,),
        (2, 3),
        (4, 5, 6),
    ],
)
def test_spherical_to_3d(shape: tuple[int, ...], xp: ArrayNamespaceFull) -> None:
    r = xp.random.random_uniform(low=0, high=1, shape=shape)
    theta = xp.random.random_uniform(low=0, high=xp.pi, shape=shape)
    phi = xp.random.random_uniform(low=0, high=2 * xp.pi, shape=shape)
    spherical: dict[Literal["r", "theta", "phi"], Array] = {
        "r": r,
        "theta": theta,
        "phi": phi,
    }
    excepted = to_cartesian(r=r, theta=theta, phi=phi)
    actual = create_spherical().to_cartesian(spherical)
    for i, value in enumerate(excepted):
        assert xp.all(xpx.isclose(value, actual[i], rtol=1e-3, atol=1e-3))  # type: ignore
    assert set(actual.keys()) == {0, 1, 2}


def to_spherical(*, x: Array, y: Array, z: Array) -> tuple[Array, Array, Array]:
    xp = array_namespace(x, y, z)
    r = xp.linalg.vector_norm(xp.stack([x, y, z], axis=0), axis=0)
    theta = xp.acos(z / r)
    phi = xp.atan2(y, x)
    return r, theta, phi


@pytest.mark.parametrize(
    "shape",
    [
        (1,),
        (2, 3),
        (4, 5, 6),
    ],
)
def test_cartesian_to_spherical_3d(
    shape: tuple[int, ...], xp: ArrayNamespaceFull
) -> None:
    x = xp.random.random_uniform(low=-1, high=1, shape=shape)
    y = xp.random.random_uniform(low=-1, high=1, shape=shape)
    z = xp.random.random_uniform(low=-1, high=1, shape=shape)
    r_expected, theta_expected, phi_expected = to_spherical(x=x, y=y, z=z)
    excepted = {"r": r_expected, "theta": theta_expected, "phi": phi_expected}
    actual = create_spherical().from_cartesian({0: x, 1: y, 2: z})
    for key, value in excepted.items():
        assert xp.all(xpx.isclose(value, actual[key], rtol=1e-3, atol=1e-3))  # type: ignore
    assert set(actual.keys()) == {"r", "theta", "phi"}


@pytest.mark.parametrize(
    "c",
    [
        create_spherical(),
        create_hopf(3),
        create_random(1),
        create_random(5),
    ],
)
@pytest.mark.parametrize(
    "shape",
    [
        (1,),
        (2, 3),
        (2, 4, 6),
    ],
)
def test_cartesian_to_spherical(
    c: SphericalCoordinates[TSpherical, TCartesian],
    shape: tuple[int, ...],
    xp: ArrayNamespaceFull,
) -> None:
    x = xp.random.random_uniform(low=-1, high=1, shape=(c.c_ndim, *shape))
    spherical = c.from_cartesian(x)
    x_reconstructed = c.to_cartesian(spherical)
    x_reconstructed = xp.stack([x_reconstructed[key] for key in range(c.c_ndim)])  # type: ignore
    assert xp.all(xpx.isclose(x, x_reconstructed, rtol=1e-6, atol=1e-6))



from typing import Literal, TypeVar

from array_api._2024_12 import Array
from array_api_compat import array_namespace, to_device
from scipy.special import hankel1, hankel2, jv, jvp, yv, yvp

TArray = TypeVar("TArray", bound=Array)


def szv(
    v: TArray,
    d: TArray,
    z: TArray,
    type: Literal["j", "y", "h1", "h2"],
    derivative: bool = False,
) -> TArray:
    r"""
    Utility function to compute hyperspherical functions.

    .. math::
        f_v^{(d)} (z) = \sqrt{\frac{\pi}{2}} \frac{F_{v + d/2 - 1}(z)}{z^{d/2 - 1}}

    Parameters
    ----------
    v : TArray
        The degree of the hyperspherical Hankel function.
    d : TArray
        The dimension of the hypersphere.
    z : TArray
        The argument of the hyperspherical Hankel function.
    type : Literal["j", "y", "h1", "h2"]
        The type of the hyperspherical function.
    derivative : bool, optional
        Whether to compute the derivative of the
        hyperspherical Hankel function, by default False

    Returns
    -------
    TArray
        The hyperspherical function.

    References
    ----------
    McLean, W. (2000). Strongly Elliptic Systems and
    Boundary Integral Equations. p.279

    """

    xp = array_namespace(v, d, z)
    if xp.any((d > 2) & (v < 0)):
        raise ValueError(
            "The hyperspherical Bessel function of "
            "the first kind is not defined for negative degrees."
        )
    if (d > 2 or type in ("h1, h2")) and derivative:
        return v / z * szv(v, d, z, type=type) - szv(v + 1, d, z, type=type)
    d_half_minus_1 = d / 2 - 1
    if type == "j":
        if derivative:
            zv = jvp
        else:
            zv = jv
    elif type == "y":
        if derivative:
            zv = yvp
        else:
            zv = yv
    elif type == "h1":
        if derivative:
            raise AssertionError()
        else:
            zv = hankel1
    elif type == "h2":
        if derivative:
            raise AssertionError()
        else:
            zv = hankel2

    dtype = xp.result_type(v, d, z, xp.float32)
    if type in ("h1", "h2"):
        dtype = xp.result_type(dtype, xp.complex64)
    return (
        xp.sqrt(xp.asarray(xp.pi / 2, device=z.device, dtype=dtype))
        * xp.asarray(
            zv(to_device(v + d_half_minus_1, "cpu"), to_device(z, "cpu")),
            device=z.device,
            dtype=dtype,
        )
        / (z**d_half_minus_1)
    )


def sjv(
    v: TArray,
    d: TArray,
    z: TArray,
    derivative: bool = False,
) -> TArray:
    r"""
    Hyperspherical Bessel function of the first kind.

    .. math::
        j_v^{(d)} (z) = \sqrt{\frac{\pi}{2}} \frac{J_{v + d/2 - 1}(z)}{z^{d/2 - 1}}

    Parameters
    ----------
    v : TArray
        The degree of the hyperspherical Bessel function.
    d : TArray
        The dimension of the hypersphere.
    z : TArray
        The argument of the hyperspherical Bessel function.
    derivative : bool, optional
        Whether to compute the derivative of the
        hyperspherical Bessel function, by default False

    Returns
    -------
    Array
        The hyperspherical Bessel function of the first kind.

    References
    ----------
    McLean, W. (2000). Strongly Elliptic Systems and
    Boundary Integral Equations. p.279

    """
    return szv(v, d, z, type="j", derivative=derivative)


def syv(
    v: TArray,
    d: TArray,
    z: TArray,
    derivative: bool = False,
) -> TArray:
    r"""
    Hyperspherical Bessel function of the second kind.

    .. math::
        y_v^{(d)} (z) = \sqrt{\frac{\pi}{2}} \frac{Y_{v + d/2 - 1}(z)}{z^{d/2 - 1}}

    Parameters
    ----------
    v : TArray
        The degree of the hyperspherical Bessel function.
    d : TArray
        The dimension of the hypersphere.
    z : TArray
        The argument of the hyperspherical Bessel function.
    derivative : bool, optional
        Whether to compute the derivative of the
        hyperspherical Bessel function, by default False

    Returns
    -------
    Array
        The hyperspherical Bessel function of the second kind.

    References
    ----------
    McLean, W. (2000). Strongly Elliptic Systems and
    Boundary Integral Equations. p.279

    """
    return szv(v, d, z, type="y", derivative=derivative)


def shn1(
    v: TArray,
    d: TArray,
    z: TArray,
    derivative: bool = False,
) -> TArray:
    r"""
    Hyperspherical Hankel function of the first kind.

    .. math::
        h_v^{(1)(d)} (z) = \sqrt{\frac{\pi}{2}} \frac{H^{(1)}_{v + d/2 - 1}(z)}{z^{d/2 - 1}}

    Parameters
    ----------
    v : TArray
        The degree of the hyperspherical Hankel function.
    d : TArray
        The dimension of the hypersphere.
    z : TArray
        The argument of the hyperspherical Hankel function.
    derivative : bool, optional
        Whether to compute the derivative of the
        hyperspherical Hankel function, by default False

    Returns
    -------
    Array
        The hyperspherical Hankel function of the first kind.

    """
    return szv(v, d, z, type="h1", derivative=derivative)


def shn2(
    v: TArray,
    d: TArray,
    z: TArray,
    derivative: bool = False,
) -> TArray:
    r"""
    Hyperspherical Hankel function of the second kind.

    .. math::
        h_v^{(2)(d)} (z) = \sqrt{\frac{\pi}{2}} \frac{H^{(2)}_{v + d/2 - 1}(z)}{z^{d/2 - 1}}

    Parameters
    ----------
    v : TArray
        The degree of the hyperspherical Hankel function.
    d : TArray
        The dimension of the hypersphere.
    z : TArray
        The argument of the hyperspherical Hankel function.
    derivative : bool, optional
        Whether to compute the derivative of the
        hyperspherical Hankel function, by default False

    Returns
    -------
    Array
        The hyperspherical Hankel function of the second kind.

    """
    return szv(v, d, z, type="h2", derivative=derivative)



from collections.abc import Sequence
from functools import lru_cache
from typing import Any, Literal

import networkx as nx
import numpy as np

from ultrasphere._coordinates import BranchingType, SphericalCoordinates
from ultrasphere._coordinates import SphericalCoordinates as cls


def _get_digraph_from_branching_type(
    branching_types: str | Sequence[BranchingType],
) -> nx.DiGraph:
    """
    Get a rooted tree from the branching types.

    Parameters
    ----------
    branching_types : str | Sequence[BranchingType]
        The branching types. e.g. "ba" for standard spherical coordinates.

    Returns
    -------
    nx.DiGraph
        The rooted tree representing the coordinates.

    Raises
    ------
    ValueError
        If the branching types are invalid.

    """
    if isinstance(branching_types, str):
        branching_types_str = branching_types.replace("bp", "b'")
        branching_types_: list[BranchingType] = []
        while branching_types_str:
            if branching_types_str.startswith("b'"):
                branching_types_.append(BranchingType.BP)
                branching_types_str = branching_types_str[2:]
            elif branching_types_str[0] in ["a", "b", "c"]:
                branching_types_.append(BranchingType(branching_types_str[0]))
                branching_types_str = branching_types_str[1:]
            else:
                raise ValueError(f"Invalid branching type: {branching_types_str}")
    else:
        branching_types_ = list(branching_types)

    G = nx.DiGraph()
    type_c_stack: list[Any] = []
    next_e_idx = 0
    next_s_idx = 0
    current_node = _s_node_name_default(next_s_idx)
    G.add_node(current_node)
    next_s_idx += 1
    for i, branching_type in enumerate(branching_types_):
        if branching_type == BranchingType.A:
            G.add_node(next_e_idx)
            G.add_edge(current_node, next_e_idx, type="cos")
            next_e_idx += 1
            G.add_node(next_e_idx)
            G.add_edge(current_node, next_e_idx, type="sin")
            next_e_idx += 1
            if i == len(branching_types_) - 1:
                break
            else:
                try:
                    next_node = type_c_stack.pop()
                except IndexError as e:
                    raise ValueError("Invalid branching types.") from e
        elif branching_type == BranchingType.B:
            G.add_node(next_e_idx)
            G.add_edge(current_node, next_e_idx, type="cos")
            next_e_idx += 1
            G.add_node(_s_node_name_default(next_s_idx))
            G.add_edge(current_node, _s_node_name_default(next_s_idx), type="sin")
            next_node = _s_node_name_default(next_s_idx)
            next_s_idx += 1
        elif branching_type == BranchingType.BP:
            G.add_node(_s_node_name_default(next_s_idx))
            G.add_edge(current_node, _s_node_name_default(next_s_idx), type="cos")
            next_node = _s_node_name_default(next_s_idx)
            next_s_idx += 1
            G.add_node(next_e_idx)
            G.add_edge(current_node, next_e_idx, type="sin")
            next_e_idx += 1
        elif branching_type == BranchingType.C:
            G.add_node(_s_node_name_default(next_s_idx))
            G.add_edge(current_node, _s_node_name_default(next_s_idx), type="cos")
            next_node = _s_node_name_default(next_s_idx)
            next_s_idx += 1
            G.add_node(_s_node_name_default(next_s_idx))
            G.add_edge(current_node, _s_node_name_default(next_s_idx), type="sin")
            type_c_stack.append(_s_node_name_default(next_s_idx))
            next_s_idx += 1
        current_node = next_node
    return G


def create_polar() -> 'SphericalCoordinates[Literal["phi"], Literal[0, 1]]':
    r"""
    Polar coordinates.

    .. math::
        x_0 &= r \cos(\phi) \\
        x_1 &= r \sin(\phi)

    Returns
    -------
    SphericalCoordinates
        The polar coordinates.

    Examples
    --------
    >>> c = create_polar()
    >>> c
    SphericalCoordinates(a)
    >>> c.s_nodes
    ['phi']
    >>> c.c_nodes
    [0, 1]

    """
    G = _get_digraph_from_branching_type("a")
    G = nx.relabel_nodes(G, {"theta0": "phi"})
    return cls(G)


def create_spherical() -> (
    'SphericalCoordinates[Literal["theta", "phi"], Literal[0, 1, 2]]'
):
    r"""
    Spherical coordinates.

    .. math::
        x_0 &= r \sin(\theta) \cos(\phi) \\
        x_1 &= r \sin(\theta) \sin(\phi) \\
        x_2 &= r \cos(\theta)

    Returns
    -------
    SphericalCoordinates
        The spherical coordinates.

    Examples
    --------
    >>> c = create_spherical()
    >>> c
    SphericalCoordinates(ba)
    >>> c.s_nodes
    ['theta', 'phi']
    >>> c.c_nodes
    [0, 1, 2]

    """
    G = _get_digraph_from_branching_type("ba")
    # swap x0 and x2
    G = nx.relabel_nodes(G, {0: 2, 2: 1, 1: 0, "theta0": "theta", "theta1": "phi"})
    return cls(G)


def create_standard(s_ndim: int) -> "SphericalCoordinates[Any, Any]":
    r"""
    Standard spherical coordinates.

    .. math::
        x_0 &= \cos(\theta_0) \\
        x_1 &= \sin(\theta_0) \cos(\theta_1) \\
        x_2 &= \sin(\theta_0) \sin(\theta_1) \cos(\theta_2) \\
        x_3 &= \sin(\theta_0) \sin(\theta_1) \sin(\theta_2) \cos(\theta_3) \\
        &\vdots \\

    Parameters
    ----------
    s_ndim : int
        The number of spherical dimensions.

    Returns
    -------
    SphericalCoordinates
        The standard coordinates.

    Examples
    --------
    >>> c = create_standard(4)
    >>> c
    SphericalCoordinates(bbba)
    >>> c.s_nodes
    ['theta0', 'theta1', 'theta2', 'theta3']
    >>> c.c_nodes
    [0, 1, 2, 3, 4]

    """
    if s_ndim == 0:
        return create_from_branching_types("")
    return cls(_get_digraph_from_branching_type("b" * (s_ndim - 1) + "a"))


def create_standard_prime(s_ndim: int) -> "SphericalCoordinates[Any, Any]":
    r"""
    Standard prime spherical coordinates.

    .. math::
        x_0 &= \sin(\theta_0) \\
        x_1 &= \cos(\theta_0) \sin(\theta_1) \\
        x_2 &= \cos(\theta_0) \cos(\theta_1) \sin(\theta_2) \\
        x_3 &= \cos(\theta_0) \cos(\theta_1) \cos(\theta_2) \sin(\theta_3) \\
        &\vdots \\

    Parameters
    ----------
    s_ndim : int
        The number of spherical dimensions.

    Returns
    -------
    SphericalCoordinates
        The standard prime coordinates.

    Examples
    --------
    >>> c = create_standard_prime(4)
    >>> c
    SphericalCoordinates(b'b'b'a)
    >>> c.s_nodes
    ['theta0', 'theta1', 'theta2', 'theta3']
    >>> c.c_nodes
    [0, 1, 2, 3, 4]

    """
    if s_ndim == 0:
        return create_from_branching_types("")
    return cls(_get_digraph_from_branching_type("bp" * (s_ndim - 1) + "a"))


def create_hopf(q: int) -> "SphericalCoordinates[Any, Any]":
    """
    Hopf coordinates.

    Parameters
    ----------
    q : int
        Where 2^q = c.c_ndim.

    Returns
    -------
    SphericalCoordinates
        The Hopf coordinates.

    Examples
    --------
    >>> c = create_hopf(3)
    >>> c
    SphericalCoordinates(ccaacaa)
    >>> c.s_nodes
    ['theta0', 'theta1', 'theta2', 'theta3', 'theta4', 'theta5', 'theta6']
    >>> c.c_nodes
    [0, 1, 2, 3, 4, 5, 6, 7]

    """

    @lru_cache
    def _hoph(q: int) -> str:
        if q < 0:
            raise ValueError("q should be non-negative.")
        elif q == 0:
            return ""
        elif q == 1:
            return "a"
        return f"c{_hoph(q - 1)}{_hoph(q - 1)}"

    return cls(_get_digraph_from_branching_type(_hoph(q)))


def create_from_branching_types(
    branching_types: str | Sequence[BranchingType],
) -> "SphericalCoordinates[Any, Any]":
    """
    Spherical coordinates from branching types.

    Parameters
    ----------
    branching_types : str | Sequence[BranchingType]
        The branching types. e.g. "ba" for standard spherical coordinates.

    Returns
    -------
    SphericalCoordinates
        The spherical coordinates.

    """
    return cls(_get_digraph_from_branching_type(branching_types))


def _s_node_name_default(idx: int) -> Any:
    """
    The naming convention for the spherical node.

    Parameters
    ----------
    idx : int
        The index of the spherical node.

    Returns
    -------
    str
        The name of the spherical node.

    """
    return f"theta{idx}"


def _e_node_name_default(idx: int) -> Any:
    """
    The naming convention for the Cartesian node.

    Parameters
    ----------
    idx : int
        The index of the Cartesian node.

    Returns
    -------
    str
        The name of the Cartesian node.

    """
    return idx


def _get_random_digraph(
    s_ndim: int, *, rng: np.random.Generator | None = None
) -> nx.DiGraph:
    """
    Get a random rooted tree representing the coordinates.

    Parameters
    ----------
    s_ndim : int
        The number of spherical dimensions.
    rng : np.random.Generator | None, optional
        The random number generator, by default None

    Returns
    -------
    nx.DiGraph
        The rooted tree representing the coordinates.

    """
    rng = np.random.default_rng() if rng is None else rng
    G = nx.DiGraph()
    leaf_nodes = [0]
    G.add_node(0)
    for _ in range(s_ndim):
        node_parent = rng.choice(leaf_nodes)
        leaf_nodes.remove(node_parent)
        node_cos = len(G)
        node_sin = len(G) + 1

        for type, node in [("cos", node_cos), ("sin", node_sin)]:
            G.add_node(node)
            G.add_edge(node_parent, node, type=type)
            leaf_nodes.append(node)
    non_leaf_nodes = set(G.nodes) - set(leaf_nodes)
    G = nx.relabel_nodes(
        G,
        {node: _e_node_name_default(i) for i, node in enumerate(leaf_nodes)}
        | {node: _s_node_name_default(i) for i, node in enumerate(non_leaf_nodes)},
    )
    return G


def create_random(
    s_ndim: int, *, rng: np.random.Generator | None = None
) -> "SphericalCoordinates[Any, Any]":
    """
    Get a random spherical coordinates.

    Parameters
    ----------
    s_ndim : int
        The number of spherical dimensions.
    rng : np.random.Generator | None, optional
        The random number generator, by default None

    Returns
    -------
    SphericalCoordinates
        The random spherical coordinates.


    Examples
    --------
    >>> from array_api_compat import numpy as np
    >>> rng = np.random.default_rng(0)
    >>> c = create_random(5, rng=rng)
    >>> c
    SphericalCoordinates(bcb'aa)
    >>> c.s_nodes
    ['theta0', 'theta1', 'theta2', 'theta3', 'theta4']
    >>> c.c_nodes
    [0, 1, 2, 3, 4, 5]

    """
    return cls(_get_random_digraph(s_ndim, rng=rng))



from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Literal, TypeVar, overload

import networkx as nx
import numpy as np
from array_api._2024_12 import Array
from array_api_compat import array_namespace
from jacobi_poly import lgamma
from strenum import StrEnum

TCartesian = TypeVar("TCartesian")
TSpherical = TypeVar("TSpherical")
COSSIN = Literal["cos", "sin"]


class BranchingType(StrEnum):
    """
    The branching types of the nodes in a rooted tree representing the coordinates.

    (Vilenkin's method of trees)
    """

    A = "a"
    B = "b"
    BP = "b'"
    C = "c"


def check_tree(graph: nx.DiGraph, /) -> None:
    """
    Check if the graph is a valid tree representing the coordinates.
    Definition of "a tree representing the coordinates":
    - The graph is a tree.
    - The graph is rooted.
    - Each node has 0 or 2 successors.
      (full binary tree, not necessarily perfect binary tree)
    - If the node has 2 successors, one of the outgoing edges has
      attribute "type" of "cos" and
      the other has attribute "type" of "sin".

    Parameters
    ----------
    graph : nx.DiGraph
        The graph representing the coordinates.

    Raises
    ------
    ValueError
        If the graph is not a valid tree.

    """
    # check if the graph is a tree
    if not nx.is_tree(graph):
        raise ValueError("The graph is not a tree.")
    # check if the graph is rooted
    if not nx.is_arborescence(graph):
        raise ValueError("The graph is not rooted.")
    # check every node has 0 or 2 successors
    for node in graph.nodes:
        if (degree := graph.out_degree(node)) not in {0, 2}:
            raise ValueError(
                f"Each node should have 0 or 2 successors, "
                f"but {node} has {degree} successors."
            )
    # check every node except the root has attribute
    # "direction" of type Literal["cos", "sin"]
    for node in graph.nodes:
        if graph.out_degree(node) == 0:
            continue
        if {e[2] for e in graph.out_edges(nbunch=node, data="type")} != {
            "cos",
            "sin",
        }:
            raise ValueError(
                f"Node {node} should have 0 or 2 successors "
                "with types 'cos' and 'sin', "
                f"got {set(graph.out_edges.data(nbunch=node, data='type'))}."
            )


def get_non_leaf_descendants(graph: nx.DiGraph, /) -> dict[Any, int]:
    """
    Calculate the number of non-leaf descendants for each node in a rooted tree.

    Parameters
    ----------
    graph : nx.DiGraph
        A rooted tree.

    Returns
    -------
    dict[Any, int]
        NaN for leaf nodes, otherwise the number of non-leaf descendants.

    """
    # calculate number of non-leaf descendants for each node
    leaf_nodes = {node for node in graph.nodes if graph.out_degree(node) == 0}
    non_leaf_descendants = {}
    for node in reversed(list(nx.topological_sort(graph))):
        if node in leaf_nodes:
            non_leaf_descendants[node] = -1
        else:
            non_leaf_descendants[node] = sum(
                non_leaf_descendants[successor] + 1
                for successor in graph.successors(node)
            )
    for node in leaf_nodes:
        non_leaf_descendants[node] = np.nan
    return non_leaf_descendants


def get_child(G: nx.DiGraph, node: Any, type: COSSIN, /) -> Any:
    """
    Get the child node of the given type in a rooted tree representing the coordinates.

    Parameters
    ----------
    G : nx.DiGraph
        A rooted tree representing the coordinates.
    node : Any
        The node to get the child.
    type : COSSIN
        The type of the child._description_

    Raises
    ------
    ValueError
        If the node has no child of the given type.
        If the node has multiple children of the given type.

    """
    child_candidates = [
        child for child in G.successors(node) if G.edges[node, child]["type"] == type
    ]
    if len(child_candidates) == 0:
        raise ValueError(f"No {type} child for node {node}.")
    elif len(child_candidates) > 1:
        raise ValueError(f"Multiple {type} children for node {node}.")
    return child_candidates[0]


def get_parent(G: nx.DiGraph, node: Any, /) -> Any | None:
    """
    Get the parent node in a rooted tree representing the coordinates.

    Parameters
    ----------
    G : nx.DiGraph
        A rooted tree representing the coordinates.
    node : Any
        The node to get the parent.

    Returns
    -------
    Any
        The parent node if exists, otherwise None.

    Raises
    ------
    ValueError
        If the node has multiple parents.

    """
    predecessors = list(G.predecessors(node))
    if len(predecessors) == 0:
        return None
    elif len(predecessors) > 1:
        raise ValueError("The node has multiple parents.")
    return predecessors[0]


def get_branching_types(G: nx.DiGraph, /) -> dict[Any, BranchingType]:
    """
    Get the branching types of each node in a rooted tree.

    Parameters
    ----------
    G : nx.DiGraph
        A rooted tree.

    Returns
    -------
    dict[Any, BranchingType]
        The branching types of each node.

    """
    branching_types = {}
    for node in G.nodes:
        if G.out_degree(node) == 0:
            continue
        sin_child = get_child(G, node, "sin")
        cos_child = get_child(G, node, "cos")
        sin_child_is_leaf = G.out_degree(sin_child) == 0
        cos_child_is_leaf = G.out_degree(cos_child) == 0
        if cos_child_is_leaf and sin_child_is_leaf:
            branching_types[node] = BranchingType.A
        elif cos_child_is_leaf and not sin_child_is_leaf:
            branching_types[node] = BranchingType.B
        elif not cos_child_is_leaf and sin_child_is_leaf:
            branching_types[node] = BranchingType.BP
        else:
            branching_types[node] = BranchingType.C
    return branching_types


def get_branching_type_from_digraph(G: nx.DiGraph, /) -> Iterable[BranchingType]:
    """
    Get the branching types of each node in a rooted tree.

    Parameters
    ----------
    G : nx.DiGraph
        A rooted tree.

    Returns
    -------
    Sequence[BranchingType]
        The branching types of each node.

    """
    root = next(nx.topological_sort(G))
    stack = [root]
    while stack:
        node = stack.pop()
        if G.out_degree(node) == 0:
            raise AssertionError()
        cos_child = get_child(G, node, "cos")
        sin_child = get_child(G, node, "sin")
        cos_child_is_leaf = G.out_degree(cos_child) == 0
        sin_child_is_leaf = G.out_degree(sin_child) == 0
        if cos_child_is_leaf and sin_child_is_leaf:
            branching_type = BranchingType.A
        elif cos_child_is_leaf and not sin_child_is_leaf:
            branching_type = BranchingType.B
            stack.append(sin_child)
        elif not cos_child_is_leaf and sin_child_is_leaf:
            branching_type = BranchingType.BP
            stack.append(cos_child)
        else:
            branching_type = BranchingType.C
            stack.extend([sin_child, cos_child])
        yield branching_type


class SphericalCoordinates[TSpherical, TCartesian]:
    """Stores the spherical coordinates using the method of trees by Vilenkin.

    Examples
    --------
    >>> import ultrasphere as us
    >>> c = us.create_standard(3)
    >>> c
    SphericalCoordinates(bba)
    >>> c.s_nodes
    ['theta0', 'theta1', 'theta2']
    >>> c.c_nodes
    [0, 1, 2, 3]
    >>> c.branching_types
    {'theta0': <BranchingType.B: 'b'>, 'theta1': <BranchingType.B: 'b'>, 'theta2': <BranchingType.A: 'a'>}
    >>> c.cos_edges
    [('theta0', 0), ('theta1', 1), ('theta2', 2)]
    >>> c.sin_edges
    [('theta0', 'theta1'), ('theta1', 'theta2'), ('theta2', 3)]
    >>> c.S
    {3: nan, 2: nan, 'theta2': 0, 1: nan, 'theta1': 1, 0: nan, 'theta0': 2}
    """

    G: nx.DiGraph
    """The rooted tree representing the coordinates."""
    S: Mapping[TSpherical, int]
    """The number of non-leaf descendants for each spherical node."""
    branching_types: Mapping[TSpherical, BranchingType]
    """The branching types of each node."""
    s_nodes: Sequence[TSpherical]
    """The spherical nodes."""
    c_nodes: Sequence[TCartesian]
    """The Cartesian nodes."""
    cos_edges: Sequence[tuple[TSpherical, TCartesian | TSpherical]]
    """The edges with type 'cos'."""
    sin_edges: Sequence[tuple[TSpherical, TCartesian | TSpherical]]
    """The edges with type 'sin'."""

    def __hash__(self) -> int:
        """__hash__."""
        return hash(self.G)

    def __getstate__(self) -> dict[Any, Any]:
        """__getstate__."""
        return {"G": nx.to_dict_of_dicts(self.G)}

    def __setstate__(self, state: dict[Any, Any]) -> None:
        """__setstate__."""
        return self.__init__(nx.from_dict_of_dicts(state["G"]))  # type: ignore

    def __eq__(self, other: Any) -> bool:
        """Check if the spherical coordinates are equal."""
        if not isinstance(other, SphericalCoordinates):
            return False
        return self.G == other.G

    def __repr__(self) -> str:
        """Print the spherical coordinates."""
        return f"SphericalCoordinates({self.branching_types_expression_str})"

    def __str__(self) -> str:
        """Print the spherical coordinates."""
        return f"SphericalCoordinates({self.branching_types_expression_str})"

    @property
    def root(self) -> Any:
        """The root node.

        Examples
        --------
        >>> import ultrasphere as us
        >>> us.create_standard(3).root
        'theta0'
        """
        return next(nx.topological_sort(self.G))

    @property
    def root_index(self) -> int:
        """The index of the root node."""
        return self.s_nodes.index(self.root)

    @property
    def branching_types_expression(self) -> Sequence[BranchingType]:
        """The branching types.

        Examples
        --------
        >>> import ultrasphere as us
        >>> us.create_standard(2).branching_types_expression
        [<BranchingType.B: 'b'>, <BranchingType.A: 'a'>]
        """
        return list(get_branching_type_from_digraph(self.G))

    @property
    def branching_types_expression_str(self) -> str:
        """The branching types as a string.

        Examples
        --------
        >>> import ultrasphere as us
        >>> us.create_standard(3).branching_types_expression_str
        'bba'
        """
        return "".join(
            branching_type.value for branching_type in self.branching_types_expression
        )

    @property
    def s_ndim(self) -> int:
        """The number of spherical dimensions.

        Examples
        --------
        >>> import ultrasphere as us
        >>> us.create_standard(3).s_ndim
        3"""
        return len(self.s_nodes)

    @property
    def c_ndim(self) -> int:
        """The number of Cartesian dimensions.

        Examples
        --------
        >>> import ultrasphere as us
        >>> us.create_standard(3).c_ndim
        4"""
        return len(self.c_nodes)

    def __init__(self, tree: nx.DiGraph, /) -> None:
        """
        Initialize the spherical coordinates.

        Parameters
        ----------
        tree : nx.DiGraph
            The rooted tree representing the coordinates.

            Definition of "a tree representing the coordinates":
            - The graph is a tree.
            - The graph is rooted.
            - Each node has 0 or 2 successors.
            (full binary tree, not necessarily perfect binary tree)
            - If the node has 2 successors, one of the outgoing edges has
            attribute "type" of "cos" and
            the other has attribute "type" of "sin".

        """
        check_tree(tree)
        self.G = tree
        self.S = get_non_leaf_descendants(tree)
        self.branching_types = get_branching_types(tree)
        nodes = list(nx.topological_sort(tree))
        # even if sorted by name, c_nodes are still topologically sorted
        # because they do not have any successors
        self.c_nodes = sorted(node for node in nodes if tree.out_degree(node) == 0)
        self.s_nodes = [node for node in nodes if tree.out_degree(node) == 2]
        self.cos_edges = [e for e in self.G.edges if self.G.edges[e]["type"] == "cos"]
        self.sin_edges = [e for e in self.G.edges if self.G.edges[e]["type"] == "sin"]

    def surface_area(self, r: float = 1) -> float:
        r"""
        The surface area of the unit sphere.

        .. math::
            |\mathbb{S}^{d-1}| = \frac{2 \pi^{d/2}}{\Gamma(d/2)} r^{d-1}

        Returns
        -------
        float
            The surface area.

        Examples
        --------
        >>> import ultrasphere as us
        >>> from array_api_compat import numpy as np
        >>> np.round(us.create_standard(2).surface_area(), 6)
        np.float64(12.566371)

        """
        return (
            2
            * (np.pi ** (self.c_ndim / 2))
            / np.exp(lgamma(self.c_ndim / 2.0))
            * r**self.s_ndim
        )

    def volume(self, r: float = 1) -> float:
        r"""
        The volume of the unit sphere.

        .. math::
            \Upsilon_d = \frac{\pi^{d/2}}{\Gamma(d/2 + 1)} r^d

        Returns
        -------
        float
            The volume.

        References
        ----------
        McLean, W. (2000). Strongly Elliptic Systems and
        Boundary Integral Equations. p.247 (Upsilon_n)

        Examples
        --------
        >>> import ultrasphere as us
        >>> from array_api_compat import numpy as np
        >>> np.round(us.create_standard(2).volume(), 6)
        np.float64(4.18879)
        """
        return (
            (np.pi ** (self.c_ndim / 2))
            / np.exp(lgamma(self.c_ndim / 2.0 + 1.0))
            * r**self.c_ndim
        )

    def from_cartesian(
        self, cartesian: Mapping[TCartesian, Array], /
    ) -> Mapping[TSpherical | Literal["r"], Array]:
        """
        Convert the Cartesian coordinates to the spherical coordinates.

        Parameters
        ----------
        cartesian : Mapping[TCartesian, Array]
            The Cartesian coordinates.

        Returns
        -------
        Array
            The spherical coordinates.

        Examples
        --------
        >>> import ultrasphere as us
        >>> from array_api_compat import numpy as np
        >>> c = us.create_spherical()
        >>> s = c.from_cartesian(np.asarray([1.0, 2.0, 3.0]))
        >>> {k: np.round(s[k], 6) for k in c.s_nodes + ["r"]}
        {'theta': np.float64(0.640522), 'phi': np.float64(1.107149), 'r': np.float64(3.741657)}

        """
        xp = array_namespace(*[cartesian[k] for k in self.c_nodes])
        r = (
            xp.linalg.vector_norm(
                xp.stack(xp.broadcast_arrays(*[cartesian[k] for k in self.c_nodes])),
                axis=0,
            )
            if self.s_ndim > 0
            else cartesian[self.c_nodes[0]]
        )
        result: dict[TSpherical | Literal["r"], Array] = {"r": r}
        tmp = {k: cartesian[k] for k in self.c_nodes}
        for node in reversed(list(nx.topological_sort(self.G))):
            if node in self.c_nodes:
                continue
            elif node not in self.s_nodes:
                raise AssertionError()
            cos_child = get_child(self.G, node, "cos")
            sin_child = get_child(self.G, node, "sin")
            cos = tmp[cos_child]
            sin = tmp[sin_child]
            result[node] = xp.atan2(sin, cos)
            tmp[node] = xp.linalg.vector_norm(
                xp.stack(xp.broadcast_arrays(sin, cos)), axis=0
            )
        return result

    @overload
    def to_cartesian(
        self,
        spherical: (
            Mapping[TSpherical | Literal["r"], Array] | Mapping[TSpherical, Array]
        ),
        /,
        *,
        as_array: Literal[False] = ...,
    ) -> Mapping[TCartesian, Array]: ...

    @overload
    def to_cartesian(
        self,
        spherical: (
            Mapping[TSpherical | Literal["r"], Array] | Mapping[TSpherical, Array]
        ),
        /,
        *,
        as_array: Literal[True] = ...,
    ) -> Array: ...

    def to_cartesian(
        self,
        spherical: (
            Mapping[TSpherical | Literal["r"], Array] | Mapping[TSpherical, Array]
        ),
        /,
        *,
        as_array: bool = False,
    ) -> Mapping[TCartesian, Array] | Array:
        """
        Convert the spherical coordinates to the Cartesian coordinates.

        Parameters
        ----------
        spherical : Mapping[TSpherical, Array]
            The spherical coordinates.
        as_array : bool, optional
            Whether to return as an array, by default False

        Returns
        -------
        Array
            The Cartesian coordinates.

        Examples
        --------
        >>> import ultrasphere as us
        >>> from array_api_compat import numpy as np
        >>> c = us.create_spherical()
        >>> e = c.to_cartesian({"theta": np.float64(0.640522), "phi": np.float64(1.107149), "r": np.float64(3.741657)})
        >>> {k: np.round(e[k], 5) for k in c.c_nodes}
        {0: np.float64(1.0), 1: np.float64(2.0), 2: np.float64(3.0)}
        """
        xp = array_namespace(*[spherical[k] for k in self.s_nodes])
        result = {self.root: spherical.get("r", 1)}  # type: ignore
        for node in nx.topological_sort(self.G):
            if node in self.c_nodes:
                continue
            elif node not in self.s_nodes:
                raise AssertionError()
            cos_child = get_child(self.G, node, "cos")
            sin_child = get_child(self.G, node, "sin")
            cos = xp.cos(spherical[node])
            sin = xp.sin(spherical[node])
            result[cos_child] = result[node] * cos
            result[sin_child] = result[node] * sin
        result = {k: result[k] for k in self.c_nodes}
        if as_array:
            return xp.stack(
                xp.broadcast_arrays(*[result[k] for k in self.c_nodes]),
                axis=0,
            )

        return result

    @property
    def is_c_keys_range(self) -> bool:
        """Whether the Cartesian keys are 0, 1, ..., c_ndim - 1."""
        return set(self.c_nodes) == set(range(self.c_ndim))

    @property
    def is_s_keys_range(self) -> bool:
        """Whether the spherical keys are 0, 1, ..., s_ndim - 1."""
        return set(self.s_nodes) == set(range(self.s_ndim))

    def jacobian(
        self,
        spherical: (
            Mapping[TSpherical | Literal["r"], Array] | Mapping[TSpherical, Array]
        ),
        /,
    ) -> Array:
        """
        Calculate the Jacobian of the spherical coordinates.

        Parameters
        ----------
        spherical : Mapping[TSpherical, Array]
            The spherical coordinates.

        Returns
        -------
        Array
            The Jacobian of the spherical coordinates.

        """
        xp = array_namespace(*[spherical[k] for k in self.s_nodes])
        if "r" in spherical:
            jacobian = spherical["r"]  # type: ignore
        else:
            jacobian = 1
        for node in self.s_nodes:
            cos = xp.cos(spherical[node])
            sin = xp.sin(spherical[node])
            cos_child = get_child(self.G, node, "cos")
            sin_child = get_child(self.G, node, "sin")
            jacobian *= cos ** (self.S.get(cos_child, -1) + 1) * sin ** (
                self.S.get(sin_child, -1) + 1
            )
        return jacobian



import matplotlib.pyplot as plt
import typer
from aquarel import load_theme

from ._creation import create_from_branching_types
from ._draw import draw

app = typer.Typer()


@app.command()
def main(branching_types: str, format: str = "jpg", theme: str = "boxy_dark") -> None:
    """
    Create a spherical coordinate system from branching types.

    Parameters
    ----------
    branching_types : str
        String representation of the branching types, e.g. "aabcc".
    format : str, optional
        The format to save the figure, by default "jpg".
    theme : str, optional
        The theme to apply to the plot, by default "boxy_dark".
        Set to "none" to disable theming.

    """
    if theme != "none":
        theme_ = load_theme(theme)
        theme_.apply()
    fig, ax = plt.subplots()
    c = create_from_branching_types(branching_types)
    draw(c, ax=ax)
    fig.savefig(f"{branching_types}.{format}")



import warnings

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.patches import Circle
from networkx.algorithms.dag import dag_longest_path
from networkx.drawing.nx_agraph import graphviz_layout

from ultrasphere._coordinates import SphericalCoordinates, TCartesian, TSpherical

_ASCII_TO_GREEK = {
    "alpha": "α",  # noqa
    "beta": "β",
    "gamma": "γ",  # noqa
    "delta": "δ",
    "epsilon": "ε",
    "zeta": "ζ",
    "theta": "θ",
    "iota": "ι",  # noqa
    "kappa": "κ",
    "lambda": "λ",
    "mu": "μ",
    "nu": "ν",  # noqa
    "xi": "ξ",
    "pi": "π",
    "rho": "ρ",  # noqa
    "sigma": "σ",  # noqa
    "tau": "τ",
    "upsilon": "υ",  # noqa
    "phi": "φ",
    "chi": "χ",
    "psi": "ψ",
    "omega": "ω",
    "eta": "η",  # last
}


def _ascii_to_greek(s: str) -> str:
    for k, v in _ASCII_TO_GREEK.items():
        s = s.replace(k, v)
    return s


def draw(
    c: SphericalCoordinates[TSpherical, TCartesian],
    root_bottom: bool = True,
    ax: Axes | None = None,
) -> tuple[float, float]:
    """
    Nicely draw the rooted tree representing the coordinates using matplotlib.

    Parameters
    ----------
    root_bottom : bool, optional
        Whether to draw the root at the bottom, by default True

    Returns
    -------
    tuple[float, float]
        The recommended width and height of the figure (in inches).

    Example
    -------
    .. skip: start

    >>> import ultrasphere as us
    >>> c = us.create_from_branching_types("ccabbab'b'ba")
    >>> us.draw(c)
    (6.5, 3.5)

    .. skip: end

    .. image:: /_static/coordinates.*

    """

    # plt.rcParams["text.usetex"] = True
    # remove spines
    ax = ax or plt.gca()
    fig = ax.figure
    ax.set_frame_on(False)
    ax.grid(False)
    width = max(c.c_ndim * 0.5 + 1, 3.5)
    height = max((len(dag_longest_path(c.G)) + 1) * 0.5, 3.5)
    additional_width = 0.8
    fig.set_size_inches(width + additional_width, height)
    fig.subplots_adjust(
        right=0.9 - additional_width / (width + additional_width),
        left=0,
        top=0.9,
        bottom=0,
    )

    # layout
    try:
        pos = graphviz_layout(c.G, prog="dot", args='-GTBbalance="max"')
        if root_bottom:
            # invert y-axis
            y_center = np.mean([y for x, y in pos.values()])
            pos = {k: (x, 2 * y_center - y) for k, (x, y) in pos.items()}
    except FileNotFoundError as e:
        warnings.warn(
            "Graphviz is not installed. "
            "The layout will be calculated by spring layout.",
            RuntimeWarning,
            stacklevel=2,
            source=e,
        )
        pos = nx.spring_layout(c.G)

    # Spherical
    nx.draw_networkx_nodes(
        c.G,
        pos,
        nodelist=c.s_nodes,
        node_color="darkgray",
        node_size=850,
        label="Spherical",
        ax=ax,
        margins=0.1,
    )
    nx.draw_networkx_labels(
        c.G,
        pos,
        labels={
            n: f"{_ascii_to_greek(str(n))}\n{c.branching_types[n].value}/{c.S[n]}"
            for n in c.s_nodes
        },
        ax=ax,
    )

    # Cartesian
    nx.draw_networkx_nodes(
        c.G,
        pos,
        nodelist=c.c_nodes,
        node_color="lightgray",
        node_shape="s",
        label="Cartesian",
        ax=ax,
    )
    nx.draw_networkx_labels(c.G, pos, labels={n: f"{n}" for n in c.c_nodes}, ax=ax)

    # edges
    cos_color = "orange"
    sin_color = "blue"
    nx.draw_networkx_edges(
        c.G,
        pos,
        edgelist=c.cos_edges,
        edge_color=cos_color,
        label="cos",
        style="dashed",
        ax=ax,
    )
    nx.draw_networkx_edges(
        c.G, pos, edgelist=c.sin_edges, edge_color=sin_color, label="sin", ax=ax
    )

    # legend
    handles = [
        Circle(
            (0, 0),
            0.25,
            facecolor="darkgray",
            label="Spherical(Name\nBranching type\n/Descendants)",
        ),
        Circle((0, 0), 0.12, facecolor="lightgray", label="Cartesian(Name)"),
        Line2D([0], [0], color=cos_color, lw=2, label="cos", linestyle="dashed"),
        Line2D([0], [0], color=sin_color, lw=2, label="sin"),
    ]
    fig.legend(
        handles=handles,
        loc="lower right" if root_bottom else "upper right",
    )
    ax.set_title(f"Type {c.branching_types_expression_str} coordinates")

    return width, height



import pytest

from ultrasphere._creation import (
    create_hopf,
    create_random,
    create_standard,
    create_standard_prime,
)


@pytest.mark.parametrize("s_ndim", [0, 1, 2, 3, 6, 103])
def test_generate(s_ndim: int) -> None:
    assert create_random(s_ndim).s_ndim == s_ndim
    assert create_standard(s_ndim).s_ndim == s_ndim
    assert create_standard_prime(s_ndim).s_ndim == s_ndim


@pytest.mark.parametrize("q", [0, 1, 2, 5])
def test_generate_hopf(q: int) -> None:
    assert create_hopf(q).c_ndim == 2**q



#!/usr/bin/env python

# This is a shim to allow GitHub to detect the package, build is done with uv
# Taken from https://github.com/Textualize/rich

import setuptools

if __name__ == "__main__":
    setuptools.setup(name="ultrasphere")



from pathlib import Path

import pytest
from matplotlib import pyplot as plt

from ultrasphere._coordinates import SphericalCoordinates
from ultrasphere._creation import (
    create_hopf,
    create_random,
    create_spherical,
    create_standard,
    create_standard_prime,
)
from ultrasphere._draw import draw

PATH = Path("tests/.cache/")
Path.mkdir(PATH, exist_ok=True)


@pytest.mark.parametrize(
    "name, c",
    [
        ("spherical", create_spherical()),
        ("hoph", create_hopf(3)),
        ("standard-4", create_standard(4)),
        ("standard-prime-4", create_standard_prime(4)),
        ("random-1", create_random(1)),
        ("random-10", create_random(10)),
    ],
)
def test_draw(name: str, c: SphericalCoordinates[str, int]) -> None:
    try:
        import pydot  # noqa
        import pygraphviz  # noqa
    except ImportError:
        pytest.skip("pydot is not installed, skipping draw test")
    draw(c)
    plt.savefig(PATH / f"{name}.jpg")
    plt.close()
